/*
 * testquaternions.cpp
 * Test routines for checking whether quaternion conversion routines work.
 *
 * Martin Wagner, March 2004
 * Distributed Wearable Augmented Reality Framework - www.augmentedreality.de
 *
 * For documentation and instructions, contact me at martin.wagner@in.tum.de.
 */

// random number generator
#include <stdlib.h>

#define TNT_BOUNDS_CHECK
#include "tnt.h"

#include "cmdlineprops.h"
#include "math_library.h"



#define STEPS 100
#define MAX_ERROR 1e-10

using namespace std;


bool checkConversionToMatrix()
{
	// create random unit quaternion
	double q[4];
	for(int i=0; i<4; i++)
		q[i] = rand();
	normalizeQuaternion(q);
	// get rotation matrix corresponding to this point
	double m[16];
	quaternionToMatrix(q,m);

	// create random 3D point
	double p[3];
	double scale = rand();
	for(int i=0; i<3; i++)
		p[i] = rand()/scale;

	double quatResult[4];
	rotateQuaternion( quatResult, q, p );
	double matResult[4];
	rotateMatrix( matResult, m, p );
	cout << "q-rot: [" << quatResult[0] << ", " << quatResult[1] << ", " << quatResult[2] << "]" << endl;
	cout << "m-rot: [" << matResult[0] << ", " << matResult[1] << ", " << matResult[2] << "]" << endl << endl;
	for(int i=0; i<3; i++)
		if( fabs(quatResult[i] - matResult[i]) >= MAX_ERROR ) return false;

	// now try converting back the matrix to a quaternion
	double q2[4];
	matrixToQuaternion(m, q2);
	for(int i=0; i<4; i++)
		if( fabs(q[i] - q2[i]) >= MAX_ERROR ) return false;

	return true;
}

bool checkMultiplication()
{
	// create random unit quaternion
	double q1[4];
	for(int i=0; i<4; i++)
		q1[i] = rand();
	normalizeQuaternion(q1);
	// get rotation matrix corresponding to this point
	double m1[16];
	quaternionToMatrix(q1,m1);

	// create second random unit quaternion
	double q2[4];
	for(int i=0; i<4; i++)
		q2[i] = rand();
	normalizeQuaternion(q2);
	// get rotation matrix corresponding to this point
	double m2[16];
	quaternionToMatrix(q2,m2);

	// create random 3D point
	double p[3];
	double scale = rand();
	for(int i=0; i<3; i++)
		p[i] = rand()/scale;

	double q3[4];
	// q3 is the quaternion representing a rotation around q1
	// followed by a rotation around q2
	multiplyQuaternion( q3, q2, q1 );

	// m3 is the corresponding matrix (first rotate around m1, then around m2)
	TNT::Array2D<double> m3 = TNT::matmult(TNT::Array2D<double>(4,4,m2), TNT::Array2D<double>(4,4,m1));
	double tmp[3];
	double twoRot[3];
	rotateQuaternion( tmp, q1, p );
	rotateQuaternion( twoRot, q2, tmp );
	double oneRot[3];
	rotateQuaternion( oneRot, q3, p );

	double matRot[3];
	rotateMatrix(matRot, m3[0], p);


	cout << "1-rot: [" << oneRot[0] << ", " << oneRot[1] << ", " << oneRot[2] << "]" << endl;
	cout << "2-rot: [" << twoRot[0] << ", " << twoRot[1] << ", " << twoRot[2] << "]" << endl;
	cout << "m-rot: [" << matRot[0] << ", " << matRot[1] << ", " << matRot[2] << "]" << endl << endl;
	for(int i=0; i<3; i++)
		if( fabs(oneRot[i] - twoRot[i]) >= MAX_ERROR || fabs(matRot[i] - twoRot[i]) >= MAX_ERROR ) return false;

	return true;
}

bool checkInversion()
{
	// create random unit quaternion
	double q[4];
	for(int i=0; i<4; i++)
		q[i] = rand();
	normalizeQuaternion(q);
	// get rotation matrix corresponding to this point
	double m[16];
	quaternionToMatrix(q,m);

	// create random 3D point
	double p[3];
	double scale = rand();
	for(int i=0; i<3; i++)
		p[i] = rand()/scale;

	// rotate point around quaternion, then around inverse
	double tmp[3];
	double quatResult[3];
	rotateQuaternion(tmp, q, p);
	rotateQuaternion(quatResult, invertQuaternion(q), tmp);

	cout << "p-orig:   [" << p[0] << ", " << p[1] << ", " << p[2] << "]" << endl;
	cout << "tmp:      [" << tmp[0] << ", " << tmp[1] << ", " << tmp[2] << "]" << endl;
	cout << "q-result: [" << quatResult[0] << ", " << quatResult[1] << ", " << quatResult[2] << "]" << endl << endl;
	for(int i=0; i<3; i++)
		if( fabs(quatResult[i] - p[i]) >= MAX_ERROR ) return false;

	return true;
}

bool checkPoseInversion()
{
	// create random unit quaternion
	double q[4];
	for(int i=0; i<4; i++)
		q[i] = rand();
	normalizeQuaternion(q);

	// create random 3D translation
	double t[3];
	double scale = rand();
	for(int i=0; i<3; i++)
		t[i] = rand()/scale;

	// create random 3D point
	double p[3];
	scale = rand();
	for(int i=0; i<3; i++)
		p[i] = rand()/scale;


	double qInv[4], tInv[3];
	invertQuaternionPose( qInv, tInv, q, t );

	// now compute pose trafo
	double pNew[3];
	rotateQuaternion( pNew, q, p );
	for(int i=0; i<3; i++) pNew[i] += t[i];

	// and compute back the original...
	double pResult[3];
	rotateQuaternion( pResult, qInv, pNew );
	for(int i=0; i<3; i++) pResult[i] += tInv[i];
	
	// and compare the results
	cout << "p-orig:   [" << p[0] << ", " << p[1] << ", " << p[2] << "]" << endl;
	cout << "p-result: [" << pResult[0] << ", " << pResult[1] << ", " << pResult[2] << "]" << endl << endl;
	for(int i=0; i<3; i++)
		if( fabs(pResult[i] - p[i]) >= MAX_ERROR ) return false;


	return true;
}


int main(int argc, char** argv)
{
	printf("%d\n", blubb());

	cmdLineProps.init(argc, argv);
	
	int steps = STEPS;
	if( cmdLineProps.find("steps") ) {
		steps = atoi(cmdLineProps.get("steps"));
	}

	// init random number generator
	struct timeval t = currentTime();
	srand(t.tv_sec + t.tv_usec);

	if( cmdLineProps.find("matConv") ) {
		for(int i=0; i<steps; i++)
			if( !checkConversionToMatrix() ) return 1;
	}
	else if( cmdLineProps.find("mult") ) {
		for(int i=0; i<steps; i++)
			if( !checkMultiplication() ) return 1;
	}
	else if( cmdLineProps.find("inv") ) {
		for(int i=0; i<steps; i++)
			if( !checkInversion() ) return 1;
	}
	else if( cmdLineProps.find("poseInv") ) {
		for(int i=0; i<steps; i++)
			if( !checkPoseInversion() ) return 1;
	}
	else {
		cerr << "No recognized option given." << endl;
		return 1;
	}

	return 0;
}


